#include <stdlib.h>

#include <string.h>
#include <vhttpd/decode.h>
#include <vhttpd/http_message.h>

#include <errno.h>

//debug
#include <stdio.h>

#include "private.h"

/*** helpers */

static const char *
_startlinetok_r(char *line, char **saveptr)
{
  return strtok_r (line, " ", saveptr);
}

static retcode_t
_parse_header (message_t * r, const char *line)
{
  char *value = NULL;

  if (strchr (_LWS, line[0]))
    {
      if (!r->lastheader_)
	return bad_request;
      value = g_strconcat (message_get_header (r, r->lastheader_),
			   " ", line + strspn (line, _LWS), NULL);
    }
  else
    {
      char *pos = strchr (line, ':');
      if (!pos)
	return bad_request;

      g_free (r->lastheader_);
      r->lastheader_ = g_ascii_strdown (line, pos - line);

      if (!*(++pos))
	return bad_request;
      value = g_strdup (pos + strspn (pos, _LWS));
    }

  message_set_header (r, r->lastheader_, value);

  g_free (value);

  return ok;
}

static const char *_COOKIE_DELIM = ";";
static const char *_COOKIE_SP = " \t=";

static void
_parse_cookies (message_t * r, const char *header)
{
  char *saveptr;
  char *result = NULL;
  char *line;

  if (!r || !header || !header[0])
    return;

  line = g_strdup (header);

  result = strtok_r (line, _COOKIE_DELIM, &saveptr);
  while (result != NULL)
    {
      char *saveptr1;
      char *lval = strtok_r (result, _COOKIE_SP, &saveptr1);
      char *rval = lval ? strtok_r (NULL, _COOKIE_SP, &saveptr1) : NULL;

      if (lval && rval)
	message_set_cookie (r, lval, rval);

      result = strtok_r (NULL, _COOKIE_DELIM, &saveptr);
    }

  g_free (line);
  return;
}

/*** functions */

int
message_read_startline_version (message_t *r,const char *version)
{
  size_t major, minor;

  if (version &&
      g_str_has_prefix (version, "HTTP/") &&
      (version += 5) &&
      (major = strspn (version, "0123456789")) &&
      (version[major] == '.') &&
      (minor = strspn (version + major + 1, "0123456789")) &&
      !version[major + minor + 1])
  {
    r->version_ = g_strdup(version);
    return 0;
  }
  return -1;
}

int
message_read_startline_method (message_t *r, const char *method)
{
  if (method &&
      method[0] &&
      (!strcmp (method, "GET") || !strcmp (method, "POST")))
  {
    message_set_method(r,method);
    return 0;
  }
  return -1;
}

int
message_read_startline_code (message_t *r, const char *code)
{
  char *err;
  size_t num;

  errno = 0;

  if (!code || !code[0])
    return -1;

  num = strtoul (code, &err, 10);

  if (!errno && !*err)
  {
    r->code_ = num;
    return 0;
  }

  return -1;
}

/* TODO: decode and split into base and query part in high level code */

int
message_read_startline_uri (message_t *r, const char *uri)
{
  if (uri && *uri)
    {
      char *pos = strchr (uri, '?');
      char *decoded_uri =  url_decode (uri, pos ? (size_t) (pos - uri) : strlen (uri));

      message_set_uri(r,decoded_uri);
      message_set_query(r,pos ? pos + 1 : "");

      g_free(decoded_uri);
      return 0;
    }
  return -1;
}

char *
message_read_line (buffer_t * buffer)
{
  char *pos;
  if (!buffer)
    return NULL;
  else if ((pos = strstr (buffer->str, _CRLF)) == NULL)
    return NULL;
  else
    {
      gsize offset = pos - buffer->str;
      char *line = g_strndup (buffer->str, offset);
      g_string_erase (buffer, (gssize) 0, (gssize) (offset + 2));
      return line;
    }
}


retcode_t
message_read_startline(message_t *r, buffer_t *buffer)
{
  char *saveptr;
  retcode_t result = ok;

  if (!buffer)
    return unfinished;

  if (r->state_ == parse_request_line)
  {
    char *line = message_read_line (buffer);
    if (!line) /* end of buffer */
	return unfinished;

    message_set_startline (r, line); /*save line before strtok*/

    switch (r->type_)
    {
	case message_response:
	    if (message_read_startline_version(r,_startlinetok_r(line,&saveptr)) ||
	        message_read_startline_code(r,_startlinetok_r(NULL,&saveptr)))
	       result = bad_request;
	    break;
	    ;;
	case message_request:
	    if (message_read_startline_method(r,_startlinetok_r(line,&saveptr)) ||
	        message_read_startline_uri(r,_startlinetok_r(NULL,&saveptr)) ||
	        message_read_startline_version(r,_startlinetok_r(NULL,&saveptr)))
	       result = bad_request;
	    break;
	    ;;
	default:
	    result = bad_request;
	    break;
	    ;;
    }

    if (result == ok)
     r->state_ = parse_headers;

    g_free(line);
  }

  return result;
}

retcode_t
message_read_header (message_t *r, buffer_t * buffer)
{
  retcode_t result = ok;

  if (!buffer)
    return unfinished;

  while (result == ok && r->state_ == parse_headers)
  {
    char *line = message_read_line (buffer);

    if (!line) /* end of buffer */
      result = unfinished;
    else if (line[0]) /* next header */
      result = _parse_header (r, line);
    else /* end of header */
    {
      _parse_cookies (r, message_get_header (r, "cookie"));
      r->state_ = parse_body;
    }

    g_free (line);
  }

  return result;
}

retcode_t
message_read_body (message_t * r, buffer_t * buffer)
{
  retcode_t result = ok;
  char *err;
  size_t c_len;
  size_t b_len;
  const char *c_len_str;

  if (!buffer)
    return unfinished;

  if(r->state_ == parse_body)
  {
    errno = 0;
    c_len_str=message_get_header (r, "content-length");
    c_len = strtoul (c_len_str?c_len_str:"", &err, 10);
    b_len = buffer->len;

    if (*err || errno)
      result = bad_request;
    else if (b_len < c_len)
      result = unfinished;
    else if (b_len > c_len)
      result = bad_request;
    else
    {
      message_set_body (r, buffer->str, b_len);
      g_string_erase (buffer, 0, -1);
    }
  }

  return result;
}

retcode_t
message_read (message_t * r, buffer_t * buffer)
{
  retcode_t result;
  if (((result = message_read_startline(r,buffer)) == ok) &&
      ((result = message_read_header(r,buffer)) == ok) &&
      ((result = message_read_body(r,buffer)) == ok))
    return ok;
  else
    return result;
}
