#define _GNU_SOURCE
#include <unistd.h>

#include <errno.h>

#include <glib.h>

#include <openssl/ssl.h>
#if OPENSSL_VERSION_NUMBER < 0x10100000L
#include <openssl/bn.h>
#include <openssl/rsa.h>
#endif


#include <vhttpd/log.h>
#include <vhttpd/tls.h>

/*** types */
struct tls_ctx_t
{
  SSL_CTX *ctx_;
};

struct tls_session_t
{
  SSL *ssl_;
  int fd_;
  int handshake_;
};

/*** helpers */

#if OPENSSL_VERSION_NUMBER < 0x10100000L
static RSA *
tmp_rsa_cb (SSL * s __attribute__ ((__unused__)),
	    int is_export __attribute__ ((__unused__)), int keylength)
{
  BIGNUM *bn = NULL;
  static RSA *rsa_tmp = NULL;

  if (!rsa_tmp && ((bn = BN_new ()) == NULL))
    runtime_error ("Allocation error in generating RSA key");
  if (!rsa_tmp && bn)
    {
      //fprintf (stderr, "Generating temp (%d bit) RSA key...\n", keylength);
      if (!BN_set_word (bn, RSA_F4) || ((rsa_tmp = RSA_new ()) == NULL) ||
	  !RSA_generate_key_ex (rsa_tmp, keylength, bn, NULL))
	{
	  if (rsa_tmp)
	    RSA_free (rsa_tmp);
	  rsa_tmp = NULL;
	}
      BN_free (bn);
    }

  return (rsa_tmp);
}
#endif

static int
tls_handshake (tls_session_t * s)
{
  int retcode = 0;
  int err;

  if (s->handshake_)
    return 0;

  retcode = SSL_accept (s->ssl_);
  err = SSL_get_error (s->ssl_, retcode);

  s->handshake_ = (retcode == 1);

  return (retcode <= 0 &&
	  err != SSL_ERROR_WANT_READ && err != SSL_ERROR_WANT_WRITE);
}


static int
tls_recv (tls_session_t * s, buffer_t *msg_buf)
{
  fd_set fds;
  char buf[BUFSIZ];
  struct timeval tv;
  int retcode = 0;

  if (!s->handshake_)
    return 0;

  do
    {
      int count = SSL_read (s->ssl_, buf, sizeof buf);

      if (count <= 0)
	{
	  int err = SSL_get_error (s->ssl_, count);
	  retcode = (err == SSL_ERROR_WANT_READ
		     || err == SSL_ERROR_WANT_WRITE) ? 0 : 1;
	  break;
	}

      g_string_append_len (msg_buf, buf, count);

      tv.tv_sec = 0;
      tv.tv_usec = 0;

      FD_ZERO (&fds);
      FD_SET (s->fd_, &fds);

    } /* First read all data from SSL buffer */
  while (SSL_pending (s->ssl_) > 0 ||
		  TEMP_FAILURE_RETRY (select (s->fd_ + 1, &fds, NULL, NULL, &tv)) == 1);

  return retcode;
}

/*** functions */

tls_ctx_t *
tls_ctx_new (const char *keyfile, const char *certfile)
{
  tls_ctx_t *c;
  const SSL_METHOD *ssl_method;

  if (!keyfile || !*keyfile)
    RUNTIME_ERROR ("keyfile is not specified");

  if (!certfile || !*certfile)
    RUNTIME_ERROR ("certfile is not specified");

  c = g_new0 (tls_ctx_t, 1);

#if OPENSSL_VERSION_NUMBER < 0x10100000L
  SSL_load_error_strings ();
  SSL_library_init ();
#else
  OPENSSL_init_ssl (0, NULL);
#endif

#if OPENSSL_VERSION_NUMBER < 0x10100000L
  ssl_method = SSLv23_server_method ();
#else
  ssl_method = TLS_server_method ();
#endif
  if (ssl_method == NULL)
    runtime_error ("Can't get TLS method");
  if ((c->ctx_ = SSL_CTX_new (ssl_method)) == NULL)
    runtime_error ("SSL_CTX_new failed");

  if (SSL_CTX_use_certificate_file (c->ctx_, certfile, SSL_FILETYPE_PEM) != 1)
    runtime_error ("error opening certificate file %s", certfile);

  if (SSL_CTX_use_PrivateKey_file (c->ctx_, keyfile, SSL_FILETYPE_PEM) != 1)
    runtime_error ("error opening private key file %s", keyfile);

  if (SSL_CTX_check_private_key (c->ctx_) != 1)
    runtime_error ("certificate doesn't match private key");

#if OPENSSL_VERSION_NUMBER < 0x10100000L
  SSL_CTX_set_tmp_rsa_callback (c->ctx_, tmp_rsa_cb);
#endif

  return c;
}

void
tls_ctx_delete (tls_ctx_t * c)
{
  if (!c)
    return;

  SSL_CTX_free (c->ctx_);

  g_free (c);
}

tls_session_t *
tls_session_new (int fd, tls_ctx_t * c)
{
  tls_session_t *s = g_new0 (tls_session_t, 1);
  s->fd_ = fd;
  s->handshake_ = FALSE;

  /* tls session init */
  if ((s->ssl_ = SSL_new (c->ctx_)) == NULL)
    runtime_error ("SSL_new failed");

  if (SSL_set_fd (s->ssl_, fd) != 1)
    runtime_error ("SSL_set_fd failed");

  return s;
}

void
tls_session_delete (tls_session_t * s)
{
  if (!s)
    return;

  do
    {
      int retcode = SSL_shutdown (s->ssl_);

      if (retcode == 0)
	{
	  int err = SSL_get_error (s->ssl_, retcode);
	  if (err == SSL_ERROR_WANT_READ || err == SSL_ERROR_WANT_WRITE)
	    continue;
	}
      break;
    }
  while (1);

  SSL_free (s->ssl_);
  g_free (s);
}


int
tls_session_send (tls_session_t * s, const char *buffer, size_t len)
{
  if (!s)
    return 1;
  do
    {
      int count = SSL_write (s->ssl_, buffer, len);

      if (count <= 0)
	{
	  int err = SSL_get_error (s->ssl_, count);
	  if (err == SSL_ERROR_WANT_READ || err == SSL_ERROR_WANT_WRITE)
	    continue;
	  else
	    return 1;
	}
      else
	{
	  buffer += count;
	  len -= count;
	};
    }
  while (len);

  return 0;
}

int
tls_session_sendfile (tls_session_t * s, int fd, size_t len)
{
  char buf[BUFSIZ];

  do
    {
      size_t rlen = MIN (len, BUFSIZ);
      ssize_t count = TEMP_FAILURE_RETRY (read (fd, buf, rlen));

      if (count < 0)
	{
	  if (errno == EAGAIN)
	    continue;
	  else
	    return 1;
	}

      if (count == 0)
	return 1;
      else
	len -= rlen;

      if (tls_session_send (s, buf, count))
	return 1;

    }
  while (len);

  return 0;
}

int
tls_session_receive (tls_session_t * s, buffer_t *buf)
{
  return !s || tls_handshake (s) || tls_recv (s, buf);
}
